#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014-2016 The Regents of the University of California,
#               Antonino Ingargiola <tritemio@gmail.com>
#
"""
The `loader` module contains functions to load each supported data format.
The loader functions load data from a specific format and
return a new :class:`fretbursts.burstlib.Data()` object containing the data.

This module contains the high-level function to load a data-file and
to return a `Data()` object. The low-level functions that perform the binary
loading and preprocessing can be found in the `dataload` folder.
"""

import os
from collections.abc import Iterable
import numpy as np
import tables

from phconvert.smreader import load_sm
from .dataload.spcreader import load_spc
from .burstlib import Data
from .utils.misc import selection_mask
from . import loader_legacy
from .burstlib_ext import group_data
import phconvert as phc

import logging
log = logging.getLogger(__name__)


def _is_multich(h5data):
    if 'photon_data' in h5data:
        return False
    elif 'photon_data0' in h5data:
        return True
    else:
        msg = 'Cannot find a photon_data group.'
        raise phc.hdf5.Invalid_PhotonHDF5(msg)


def _append_data_ch(d, name, value):
    if name not in d:
        d.add(**{name: [value]})
    else:
        d[name].append(value)


def _load_from_group(d, group, name, dest_name, multich_field=False,
                     ondisk=False, allow_missing=True):
    if allow_missing and name not in group:
        return

    node_value = group._f_get_child(name)
    if not ondisk:
        node_value = node_value.read()
    if multich_field:
        _append_data_ch(d, dest_name, node_value)
    else:
        d.add(**{dest_name: node_value})


def _append_empy_ch(data):
    # Empty channel, fill it with empty arrays
    ph_times = np.array([], dtype='int64')
    _append_data_ch(data, 'ph_times_m', ph_times)

    a_em = np.array([], dtype=bool)
    _append_data_ch(data, 'A_em', a_em)


def _get_measurement_specs(ph_data, setup):
    if 'measurement_specs' not in ph_data:
        # No measurement specs, we will load timestamps and set them all in a
        # conventional photon stream (acceptor emission)
        meas_type = 'smFRET-1color'
        meas_specs = None
    else:
        assert 'measurement_type' in ph_data.measurement_specs
        meas_specs = ph_data.measurement_specs
        meas_type = meas_specs.measurement_type.read().decode()

    if meas_type not in ['smFRET-1color', 'smFRET',
                         'smFRET-usALEX', 'smFRET-nsALEX', 'generic']:
        raise NotImplementedError('Measurement type "%s" not supported'
                                  ' by FRETBursts.' % meas_type)
    num_spectral_ch = setup.num_spectral_ch.read()
    num_polarization_ch = setup.num_polarization_ch.read()
    num_split_ch = setup.num_split_ch.read()
    if meas_type == 'generic':
        msg = ('This file contains {n} {type} channels.\n'
               'Unfortunately, the current FRETBursts version only supports\n'
               '{nvalid} {type} channel.')
        if num_polarization_ch > 2:
            raise ValueError(msg.format(n=num_polarization_ch,
                                        type='polarization', nvalid='1 or 2'))
        if num_spectral_ch > 2:
            raise ValueError(msg.format(n=num_spectral_ch,
                                        type='spectral', nvalid='1 or 2'))
        if num_split_ch == 2 and num_spectral_ch == 1:
            # in this case data will be loaded as "spectral".
            log.warning('Loading split channels as spectral channels.')
        elif num_split_ch != 1:
            raise ValueError(msg.format(n=setup.num_split_ch, type='split',
                                        nvalid=1))

        if num_spectral_ch == 1 and num_split_ch == 1:
            # One laser one detector
            meas_type = 'smFRET-1color'
        elif not setup.modulated_excitation.read():
            # Single laser and two spectral (or split) detection channels
            meas_type = 'smFRET'
        elif tuple(setup.excitation_alternated.read()) == (False, True):
            meas_type = 'PAX'
        elif tuple(setup.excitation_alternated.read()) == (True, True):
            if setup.lifetime.read():
                meas_type = 'smFRET-nsALEX'
            else:
                meas_type = 'smFRET-usALEX'
        if setup.num_polarization_ch.read() > 1:
            meas_type += '-2pol'
        # Check consistency of polarization specs
        if meas_specs is not None:
            det_specs = meas_specs.detectors_specs
            if setup.num_polarization_ch.read() == 1:
                if (
                    'polarization_ch1' in det_specs
                    and 'polarization_ch2' in det_specs
                ):
                    msg = ("The field `/setup/num_polarization_ch` indicates "
                           "no polarization.\nHowever, the fields "
                           "`detectors_specs/polarization_ch*` are present.")
                    raise phc.hdf5.Invalid_PhotonHDF5(msg)
            else:
                if not (
                    'polarization_ch1' in det_specs
                    and 'polarization_ch2' in det_specs
                ):
                    msg = ("The field `/setup/num_polarization_ch` indicates "
                           "more than one polarization.\nHowever, some "
                           "`detectors_specs/polarization_ch*` fields are "
                           "missing.")
                    raise phc.hdf5.Invalid_PhotonHDF5(msg)
    return meas_type, meas_specs


def _load_photon_data_arrays(data, ph_data, ondisk=False):
    assert 'timestamps' in ph_data

    # Build mapping to convert Photon-HDF5 to FRETBursts names
    # fields not mapped use the same name on both Photon-HDF5 and FRETBursts
    mapping = {'timestamps': 'ph_times_m',
               'nanotimes': 'nanotimes', 'particles': 'particles'}
    if data.alternated:
        mapping = {'timestamps': 'ph_times_t', 'detectors': 'det_t',
                   'nanotimes': 'nanotimes_t', 'particles': 'particles_t'}

    # Load all photon-data arrays
    for name in ph_data._v_leaves:
        dest_name = mapping.get(name, name)
        _load_from_group(data, ph_data, name, dest_name=dest_name,
                         multich_field=True, ondisk=ondisk)

    # Timestamps are always present, and their units are always present too
    data.add(clk_p=ph_data.timestamps_specs.timestamps_unit.read())


def _load_nanotimes_specs(data, ph_data):
    nanot_specs = ph_data.nanotimes_specs
    nanotimes_params = {}
    for name in ['tcspc_unit', 'tcspc_num_bins', 'tcspc_range']:
        value = nanot_specs._f_get_child(name).read()
        nanotimes_params.update(**{name: value})
    if 'user' in nanot_specs:
        for name in ['tau_accept_only', 'tau_donor_only',
                     'tau_fret_donor', 'inverse_fret_rate']:
            if name in nanot_specs.user:
                value = nanot_specs.user._f_get_child(name).read()
                nanotimes_params.update(**{name: value})
    _append_data_ch(data, 'nanotimes_params', nanotimes_params)


def _add_usALEX_specs(data, meas_specs):
    # Used for both us-ALEX and PAX
    try:
        offset = meas_specs.alex_offset.read()
    except tables.NoSuchNodeError:
        log.warning('    No offset found, assuming offset = 0.')
        offset = 0
    data.add(offset=offset)
    data.add(alex_period=meas_specs.alex_period.read())
    _load_alex_periods_donor_acceptor(data, meas_specs)


def _load_alex_periods_donor_acceptor(data, meas_specs):
    # Used for both us- and ns-ALEX and PAX
    try:
        # Try to load alex period definitions
        D_ON = meas_specs.alex_excitation_period1.read()
        A_ON = meas_specs.alex_excitation_period2.read()
    except tables.NoSuchNodeError:
        # But if it fails it's OK, those fields are optional
        msg = """
        The current file lacks the alternation period definition.
        You will need to manually add this info using:

          d.add(D_ON=D_ON, A_ON=A_ON)

        where `d` is a Data object and D_ON/A_ON is a tuple with start/stop
        values defining the D/A excitation excitation period. Values are in
        raw timestamps units.
        """
        log.warning(msg)
    else:
        data.add(D_ON=D_ON, A_ON=A_ON)


def _compute_acceptor_emission_mask(data, ich, ondisk):
    """For non-ALEX measurements."""
    if data.detectors[ich].dtype.itemsize != 1:
        raise NotImplementedError('Detectors dtype must be 1-byte.')
    donor, accept = data._det_donor_accept_multich[ich]

    # Remove counts not associated with D or A channels
    det_ich = data.detectors[ich][:]  # load the data in case ondisk = True
    num_detectors = len(np.unique(det_ich))
    if not ondisk and num_detectors > donor.size + accept.size:
        mask = (selection_mask(det_ich, donor) +
                selection_mask(det_ich, accept))
        data.detectors[ich] = det_ich[mask]
        data.ph_times_m[ich] = det_ich[mask]
        if 'nanotimes' in data:
            data.nanotimes[ich] = data.nanotimes[ich][:][mask]

    # From `detectors` compute boolean mask `A_em`
    if not ondisk and donor.size == 1 and 0 in (accept, donor):
        # In this case we create the boolean mask in-place
        # using the detectors array
        _append_data_ch(data, 'A_em', data.detectors[ich].view(dtype=bool))
        if accept == 0:
            np.logical_not(data.A_em[ich], out=data.A_em[ich])
    else:
        # Create the boolean mask as a new array
        _append_data_ch(data, 'A_em',
                        selection_mask(det_ich, accept))


def _photon_hdf5_1ch(h5data, data, ondisk=False, nch=1, ich=0, loadspecs=True):
    data.add(nch=nch)
    ph_data_name = '/photon_data' if nch == 1 else '/photon_data%d' % ich

    # Handle the case of missing channel (e.g. dead pixel)
    if ph_data_name not in h5data:
        _append_empy_ch(data)
        return

    # Load photon_data group and measurement_specs (if present)
    ph_data = h5data._f_get_child(ph_data_name)
    meas_type, meas_specs = _get_measurement_specs(ph_data, h5data.setup)
    # Set some `data` flags
    data.add(meas_type=meas_type)
    data.add(ALEX='ALEX' in meas_type)  # True for usALEX, nsALEX and PAX
    data.add(alternated=data.ALEX or 'PAX' in data.meas_type)
    data.add(lifetime='nanotimes' in ph_data)
    data.add(polarization='2pol' in meas_type)
    data.add(spectral='smFRET-1color' not in meas_type)

    # Load photon_data arrays
    _load_photon_data_arrays(data, ph_data, ondisk=ondisk)

    # If nanotimes are present load their specs
    if data.lifetime:
        _load_nanotimes_specs(data, ph_data)

    # Unless 1-color, load donor and acceptor info
    det_specs = meas_specs.detectors_specs
    if data.spectral:
        try:
            donor = np.atleast_1d(det_specs.spectral_ch1.read())
            accept = np.atleast_1d(det_specs.spectral_ch2.read())
        except tables.NoSuchNodeError:
            donor = np.atleast_1d(det_specs.split_ch1.read())
            accept = np.atleast_1d(det_specs.split_ch2.read())
        _append_data_ch(data, 'det_donor_accept', (donor, accept))
    else:
        # Non-FRET or unspecified data, assume all photons are "acceptor"
        _append_data_ch(data, 'A_em', slice(None))

    if data.polarization:
        p_pol = np.atleast_1d(det_specs.polarization_ch1.read())
        s_pol = np.atleast_1d(det_specs.polarization_ch2.read())
        _append_data_ch(data, 'det_p_s_pol', (p_pol, s_pol))

    # Here there are all the special-case for each measurement type
    if data.spectral and not data.alternated:
        # No alternation, we can compute the emission masks right away
        _compute_acceptor_emission_mask(data, ich, ondisk=ondisk)

    if loadspecs and data.spectral and data.alternated and not data.lifetime:
        # load alternation metadata for usALEX or PAX
        _add_usALEX_specs(data, meas_specs)

    if loadspecs and data.lifetime:
        data.add(laser_repetition_rate=meas_specs.laser_repetition_rate.read())
        if data.ALEX:
            # load alternation metadata for nsALEX
            _load_alex_periods_donor_acceptor(data, meas_specs)


def _photon_hdf5_multich(h5data, data, ondisk=True):
    ph_times_dict = phc.hdf5.photon_data_mapping(h5data._v_file)
    nch = np.max(list(ph_times_dict.keys())) + 1
    _photon_hdf5_1ch(h5data, data, ondisk=ondisk, nch=nch, ich=0)
    for ich in range(1, nch):
        _photon_hdf5_1ch(h5data, data, ondisk=ondisk, nch=nch, ich=ich,
                         loadspecs=False)


def photon_hdf5(filename, ondisk=False, require_setup=True, validate=False, fix_order=True):
    """Load a data file saved in Photon-HDF5 format version 0.3 or higher.

    Photon-HDF5 is a format for a wide range of timestamp-based
    single molecule data. For more info please see:

    http://photon-hdf5.org/

    Arguments:
        filename (str or pathlib.Path): path of the data file to be loaded.
        ondisk (bool): if True, do not load the timestamps in memory
            using instead references to the HDF5 arrays. Default False.
        require_setup (bool): if True (default) the input file need to
            have a setup group or won't be loaded. If False, accept files
            with missing setup group. Use False only for testing or
            DCR files.
        validate (bool): if True validate the Photon-HDF5 file on loading.
            If False skip any validation.
        fix_order (bool): if True fix the order of photons so all are monotonically
            increaseing in macrotime.

    Returns:
        :class:`fretbursts.burstlib.Data` object containing the data.
    """
    if not isinstance(filename, str) and isinstance(filename, Iterable):
        d = group_data([photon_hdf5(f) for f in filename])
        return d
    filename = str(filename)
    assert os.path.isfile(filename), 'File not found.'
    version = phc.hdf5._check_version(filename)
    if version == u'0.2':
        return loader_legacy.hdf5(filename)

    h5file = tables.open_file(filename)
    try:
        # make sure the file is valid
        if validate and version.startswith(u'0.4'):
            phc.v04.hdf5.assert_valid_photon_hdf5(h5file,
                                                  require_setup=require_setup,
                                                  strict_description=False)
        elif validate:
            phc.hdf5.assert_valid_photon_hdf5(h5file, require_setup=require_setup,
                                              strict_description=False)
        # Create the data container
        h5data = h5file.root
        d = Data(fname=filename, data_file=h5data._v_file)
    
        for grp_name in ['setup', 'sample', 'provenance', 'identity']:
            if grp_name in h5data:
                d.add(**{grp_name:
                         phc.hdf5.dict_from_group(h5data._f_get_child(grp_name))})
    
        for field_name in ['description', 'acquisition_duration']:
            if field_name in h5data:
                d.add(**{field_name: h5data._f_get_child(field_name).read()})
    
        if _is_multich(h5data):
            _photon_hdf5_multich(h5data, d, ondisk=ondisk)
        else:
            _photon_hdf5_1ch(h5data, d, ondisk=ondisk)
        if fix_order:
            sort_photon_times(d)
    finally:
        if not ondisk:
            h5file.close()
    return d


##
# Multi-spot loader functions
#

##
# usALEX loader functions
#

# Build masks for the alternating periods
def _select_outer_range(times, period, edges):
    return ((times % period) >= edges[0]) + ((times % period) < edges[1])


def _select_inner_range(times, period, edges):
    return ((times % period) >= edges[0]) * ((times % period) < edges[1])


def _select_range(times, period, edges):
    return _select_inner_range(times, period, edges) if edges[0] < edges[1] \
        else _select_outer_range(times, period, edges)


def usalex(fname, leakage=0, gamma=1., header=None, BT=None):
    """Load usALEX data from a SM file and return a Data() object.

    This function returns a Data() object to which you need to apply
    an alternation selection before performing further analysis (background
    estimation, burst search, etc.).

    The pattern to load usALEX data is the following::

        d = loader.usalex(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580), alex_period=4000)
        plot_alternation_hist(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for further processing such as background estimation,
    burst search, etc...
    """
    if BT is not None:
        log.warning('`BT` argument is deprecated, use `leakage` instead.')
        leakage = BT
    if header is not None:
        log.warning('    `header` argument ignored. '
                    '    The header length is now computed automatically.')
    print(" - Loading '%s' ... " % fname)
    ph_times_t, det_t, labels = load_sm(fname, return_labels=True)
    print(" [DONE]\n")

    DONOR_ON = (2850, 580)
    ACCEPT_ON = (930, 2580)
    alex_period = 4000

    dx = Data(fname=fname, clk_p=12.5e-9, nch=1, leakage=leakage, gamma=gamma,
              ALEX=True, lifetime=False, alternated=True,
              meas_type='smFRET-usALEX', polarization=False,
              D_ON=DONOR_ON, A_ON=ACCEPT_ON, alex_period=alex_period,
              ph_times_t=[ph_times_t], det_t=[det_t],
              det_donor_accept=(np.atleast_1d(0), np.atleast_1d(1)),
              ch_labels=labels)
    return dx


def _usalex_apply_period_1ch(d, delete_ph_t=True, remove_d_em_a_ex=False,
                             ich=0):
    """Applies to the Data object `d` the alternation period previously set.

    This function operates on a single-channel.
    See :func:`usalex_apply_period` for details.
    """
    donor_ch, accept_ch = d._det_donor_accept_multich[ich]
    D_ON, A_ON = d._D_ON_multich[ich], d._A_ON_multich[ich]
    # Remove eventual ch different from donor or acceptor
    det_t = d.det_t[ich][:]
    ph_times_t = d.ph_times_t[ich][:]
    d_ch_mask_t = selection_mask(det_t, donor_ch)
    a_ch_mask_t = selection_mask(det_t, accept_ch)
    valid_det = d_ch_mask_t + a_ch_mask_t

    # Build masks for excitation windows
    if 'offset' in d:
        ph_times_t -= d.offset
    d_ex_mask_t = _select_range(ph_times_t, d.alex_period, D_ON)
    a_ex_mask_t = _select_range(ph_times_t, d.alex_period, A_ON)
    # Safety check: each ph is either D or A ex (not both)
    assert not (d_ex_mask_t * a_ex_mask_t).any()

    # Select alternation periods, removing transients and invalid detectors
    DexAex_mask = (d_ex_mask_t + a_ex_mask_t) * valid_det

    # Reduce photons to the DexAex_mask selection
    ph_times = ph_times_t[DexAex_mask]
    d_em = d_ch_mask_t[DexAex_mask]
    a_em = a_ch_mask_t[DexAex_mask]
    d_ex = d_ex_mask_t[DexAex_mask]
    a_ex = a_ex_mask_t[DexAex_mask]
    assert d_ex.sum() == d_ex_mask_t.sum()
    assert a_ex.sum() == a_ex_mask_t.sum()

    if remove_d_em_a_ex:
        # Removes donor-ch photons during acceptor excitation
        mask = a_em + d_em * d_ex
        assert (mask == -(a_ex * d_em)).all()
        ph_times = ph_times[mask]
        d_em = d_em[mask]
        a_em = a_em[mask]
        d_ex = d_ex[mask]
        a_ex = a_ex[mask]

    assert d_em.sum() + a_em.sum() == ph_times.size
    assert (d_em + a_em).all()       # masks fill the total array
    assert not (d_em * a_em).any()   # no photon is both D and A
    assert a_ex.size == a_em.size == d_ex.size == d_em.size == ph_times.size
    _append_data_ch(d, 'ph_times_m', ph_times)
    _append_data_ch(d, 'D_em', d_em)
    _append_data_ch(d, 'A_em', a_em)
    _append_data_ch(d, 'D_ex', d_ex)
    _append_data_ch(d, 'A_ex', a_ex)
    assert (len(d.ph_times_m) == len(d.D_em) == len(d.A_em) ==
            len(d.D_ex) == len(d.A_ex) == ich + 1)

    if 'particles_t' in d:
        particles_t = d.particles_t[ich][:]
        particles = particles_t[DexAex_mask]
        _append_data_ch(d, 'particles', particles)

    assert d.ph_times_m[ich].size == d.A_em[ich].size

    if d.polarization:
        # We also have polarization data
        p_pol_ch, s_pol_ch = d._det_p_s_pol_multich[ich]
        p_em, s_em = _get_det_masks(det_t, p_pol_ch, s_pol_ch, DexAex_mask,
                                    mask_ref=valid_det, ich=ich)
        _append_data_ch(d, 'P_em', p_em)
        _append_data_ch(d, 'S_em', s_em)

    if delete_ph_t:
        d.delete('ph_times_t')
        d.delete('det_t')
    return d


def usalex_apply_period(d, delete_ph_t=True, remove_d_em_a_ex=False):
    """Applies to the Data object `d` the alternation period previously set.

    Note that you first need to load the data in a variable `d` and then
    set the alternation parameters using `d.add(D_ON=..., A_ON=...)`.

    The typical pattern for loading ALEX data is the following::

        d = loader.photon_hdf5(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for further processing such as background estimation,
    burst search, etc...

    *See also:* :func:`alex_apply_period`.
    """
    for ich in range(d.nch):
        _usalex_apply_period_1ch(d, remove_d_em_a_ex=remove_d_em_a_ex, ich=ich,
                                 delete_ph_t=False)
    if delete_ph_t:
        d.delete('ph_times_t')
        d.delete('det_t')
    d.add(alternation_applied=True)
    return d

##
# nsALEX loader functions
#

def nsalex(fname):
    """Load nsALEX data from a SPC file and return a Data() object.

    This function returns a Data() object to which you need to apply
    an alternation selection before performing further analysis (background
    estimation, burst search, etc.).

    The pattern to load nsALEX data is the following::

        d = loader.nsalex(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for further processing such as background estimation,
    burst search, etc...
    """
    ph_times_t, det_t, nanotimes = load_spc(fname)

    DONOR_ON = (10, 1500)
    ACCEPT_ON = (2000, 3500)
    nanotimes_nbins = 4095

    dx = Data(fname=fname, clk_p=50e-9, nch=1, ALEX=True, lifetime=True,
              D_ON=DONOR_ON, A_ON=ACCEPT_ON,
              nanotimes_nbins=nanotimes_nbins,
              nanotimes_params=[{'tcspc_num_bins': nanotimes_nbins}],
              ph_times_t=[ph_times_t], det_t=[det_t], nanotimes_t=[nanotimes],
              det_donor_accept=(np.atleast_1d(4), np.atleast_1d(6)))
    return dx


def _nsalex_apply_period_1ch(d, i, delete_ph_t=True):
    """Applies to the Data object `d` the alternation period previously set.

    Note that you first need to load the data in a variable `d` and then
    set the alternation parameters using `d.add(D_ON=..., A_ON=...)`.

    The typical pattern for loading ALEX data is the following::

        d = loader.photon_hdf5(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for further processing such as background estimation,
    burst search, etc...

    *See also:* :func:`alex_apply_period`.
    """
    ich = i
    donor_ch, accept_ch = d._det_donor_accept_multich[ich]
    D_ON_multi, A_ON_multi = d._D_ON_multich[ich], d._A_ON_multich[ich]
    D_ON = [(D_ON_multi[i], D_ON_multi[i + 1])
            for i in range(0, len(D_ON_multi), 2)]
    A_ON = [(A_ON_multi[i], A_ON_multi[i + 1])
            for i in range(0, len(A_ON_multi), 2)]
    # Mask for donor + acceptor detectors (discard other detectors)
    det_t = d.det_t[ich][:]
    d_ch_mask_t = selection_mask(det_t, donor_ch)
    a_ch_mask_t = selection_mask(det_t, accept_ch)
    da_ch_mask_t = d_ch_mask_t + a_ch_mask_t

    # Masks for excitation periods
    d_ex_mask_t = np.zeros(d.nanotimes_t[ich].size, dtype='bool')
    for d_on in D_ON:
        d_ex_mask_t += (d.nanotimes_t[ich] > d_on[0]) * (d.nanotimes_t[ich] < d_on[1])

    a_ex_mask_t = np.zeros(d.nanotimes_t[ich].size, dtype='bool')
    for a_on in A_ON:
        a_ex_mask_t += (d.nanotimes_t[ich] > a_on[0]) * (d.nanotimes_t[ich] < a_on[1])

    ex_mask_t = d_ex_mask_t + a_ex_mask_t  # Select only ph during Dex or Aex

    # Total mask: D+A photons, and only during the excitation periods
    valid = da_ch_mask_t * ex_mask_t  # logical AND

    # Apply selection to timestamps and nanotimes
    ph_times = d.ph_times_t[ich][:][valid]
    nanotimes = d.nanotimes_t[ich][:][valid]

    # Apply selection to the emission masks
    d_em = d_ch_mask_t[valid]
    a_em = a_ch_mask_t[valid]
    assert (d_em + a_em).all()       # masks fill the total array
    assert not (d_em * a_em).any()   # no photon is both D and A

    # Apply selection to the excitation masks
    d_ex = d_ex_mask_t[valid]
    a_ex = a_ex_mask_t[valid]
    assert (d_ex + a_ex).all()
    assert not (d_ex * a_ex).any()
    _append_data_ch(d, 'ph_times_m', ph_times)
    _append_data_ch(d, 'nanotimes', nanotimes)
    _append_data_ch(d, 'D_em', d_em)
    _append_data_ch(d, 'A_em', a_em)
    _append_data_ch(d, 'D_ex', d_ex)
    _append_data_ch(d, 'A_ex', a_ex)
    
    if d.polarization:
        # We also have polarization data
        p_polariz_ch, s_polariz_ch = d._det_p_s_pol_multich[ich]
        p_em, s_em = _get_det_masks(det_t, p_polariz_ch, s_polariz_ch, valid,
                                    ich=ich)
        _append_data_ch(d, 'P_em', p_em)
        _append_data_ch(d, 'S_em', s_em)

def nsalex_apply_period(d, delete_ph_t=True):
    """Applies to the Data object `d` the alternation period previously set.

    Note that you first need to load the data in a variable `d` and then
    set the alternation parameters using `d.add(D_ON=..., A_ON=...)`.

    The typical pattern for loading ALEX data is the following::

        d = loader.photon_hdf5(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for further processing such as background estimation,
    burst search, etc...

    *See also:* :func:`alex_apply_period`.
    """
    for ich in range(d.nch):
        _nsalex_apply_period_1ch(d,ich)
    if delete_ph_t:
        d.delete('ph_times_t')
        d.delete('det_t')
        d.delete('nanotimes_t')
    d.add(alternation_applied=True)
    return d


def _get_det_masks(det_t, det_ch1, det_ch2, valid, mask_ref=None, ich=0):
    """
    Returns two masks for photons detected by `det_ch1` and `det_ch2`,
    and being also `valid`. `valid` is a bool mask, same size as `det_t`,
    which selects the photons considered "valid".
    The returned masks have same size as `det_t` (and as `valid`).
    """
    ch1_mask_t = selection_mask(det_t, det_ch1)
    ch2_mask_t = selection_mask(det_t, det_ch2)
    if mask_ref is not None:
        both_ch_mask_t = ch1_mask_t + ch2_mask_t
        assert all(both_ch_mask_t == mask_ref)
    # Apply "valid" selection to the channel masks
    ch1_mask = ch1_mask_t[valid]
    ch2_mask = ch2_mask_t[valid]
    assert (ch1_mask + ch2_mask).all()       # masks fill the total array
    assert not (ch1_mask * ch2_mask).any()   # no photon is both channels
    return ch1_mask, ch2_mask


def alex_apply_period(d, delete_ph_t=True):
    """Apply the ALEX period definition set in D_ON and A_ON attributes.

    This function works both for us-ALEX and ns-ALEX data.

    Note that you first need to load the data in a variable `d` and then
    set the alternation parameters using `d.add(D_ON=..., A_ON=...)`.

    The typical pattern for loading ALEX data is the following::

        d = loader.photon_hdf5(fname=fname)
        d.add(D_ON=(2850, 580), A_ON=(900, 2580))
        alex_plot_alternation(d)

    If the plot looks good, apply the alternation with::

        loader.alex_apply_period(d)

    Now `d` is ready for further processing such as background estimation,
    burst search, etc...
    """
    if not d.alternated:
        print('No alternation found. Nothing to apply.')
        return
    if 'alternation_applied' in d and d['alternation_applied']:
        print('Alternation already applied, I cannot reapply it. \n'
              'Reload the data if you need to change alternation parameters.')
        return

    if d.lifetime:
        apply_period_func = nsalex_apply_period
    else:
        apply_period_func = usalex_apply_period
    apply_period_func(d, delete_ph_t=delete_ph_t)
    msg = ('# Total photons (after ALEX selection):  {:12,}\n'
           '#  D  photons in D+A excitation periods: {:12,}\n'
           '#  A  photons in D+A excitation periods: {:12,}\n'
           '# D+A photons in  D  excitation period:  {:12,}\n'
           '# D+A photons in  A  excitation period:  {:12,}\n')
    ph_data_size = d.ph_data_sizes.sum()
    D_em_sum = sum(d_em.sum() for d_em in d.D_em)
    A_em_sum = sum(a_em.sum() for a_em in d.A_em)
    D_ex_sum = sum(d_ex.sum() for d_ex in d.D_ex)
    A_ex_sum = sum(a_ex.sum() for a_ex in d.A_ex)
    print(msg.format(ph_data_size, D_em_sum, A_em_sum, D_ex_sum, A_ex_sum))


def sm_single_laser(fname):
    """Load SM files acquired using single-laser and 2 detectors.
    """
    print(" - Loading '%s' ... " % fname)
    ph_times_t, det_t, labels = load_sm(fname, return_labels=True)
    print(" [DONE]\n")

    a_em = (det_t == 1)
    dx = Data(fname=fname, clk_p=12.5e-9, nch=1,
              ALEX=False, lifetime=False, alternated=False,
              meas_type='smFRET',
              ph_times_m=[ph_times_t], det_donor_accept=(0, 1), A_em=[a_em],
              ch_labels=labels)
    dx.add(acquisition_duration=np.round(dx.time_max - dx.time_min, 1))
    return dx

def sort_photon_times(d):
    for i in range(d.nch):
        if hasattr(d, 'ph_times_t'):
            sort_t = np.argsort(d.ph_times_t[i])
            d.ph_times_t[i] = d.ph_times_t[i][sort_t]
            d.det_t[i] = d.det_t[i][sort_t]
            if hasattr(d, 'nanotimes_t'):
                d.nanotimes_t[i] = d.nanotimes_t[i][sort_t]
        if hasattr(d, 'ph_times_m'):
            sort_m = np.argsort(d.ph_times_m[i])
            d.ph_times_m[i] = d.ph_times_m[i][sort_m]
            if hasattr(d, 'nanotimes'):
                d.nanotimes[i] = d.nanotimes[i][sort_m]
            if hasattr(d, 'A_em'):
                d.A_em[i] = d.A_em[i][sort_m]
            if hasattr(d, 'D_em'):
                d.D_em[i] = d.D_em[i][sort_m]
            if hasattr(d, 'A_ex'):
                d.A_ex[i] = d.A_ex[i][sort_m]
            if hasattr(d, 'D_ex'):
                d.D_ex[i] = d.D_ex[i][sort_m]


"""
Example on how to use Matplotlib events to allow the user to select a region
of the plot.
"""


from scipy import *
from matplotlib.patches import Rectangle, Ellipse

xs, ys, xe, ye = 0,0,0,0
id_motion = 0
r = 0

Shape = Rectangle
#Shape = Ellipse

def on_press(event):
    global id_motion, xs, ys, r
    print 'INAXES: ', event.inaxes
    if event.inaxes!=ax: return
    xs, ys = event.xdata, event.ydata
    print 'PRESS button=%d, x=%d, y=%d, xdata=%f, ydata=%f' % (
            event.button, event.x, event.y, event.xdata, event.ydata)
    r = Rectangle(xy=(xs,ys), height=0, width=0, fill=False, lw=2, alpha=0.2)
    ax.add_artist(r)
    r.set_clip_box(ax.bbox)
    draw()
    id_motion = fig.canvas.mpl_connect('motion_notify_event', on_motion)

def on_release(event):
    #print 'RELEASE button=%d, x=%d, y=%d, xdata=%f, ydata=%f' % (
    #        event.button, event.x, event.y, event.xdata, event.ydata)
    print "Selection: ", xs, ys, xe, ye
    fig.canvas.mpl_disconnect(id_motion)

def on_motion(event):
    global xe, ye
    #print 'MOTION x=%d, y=%d, xdata=%f, ydata=%f' % (
    #        event.x, event.y, event.xdata, event.ydata)
    xe, ye = event.xdata, event.ydata
    print xe,ye
    r.set_height(ye-ys)
    r.set_width(xe-xs)
    draw()

sx = normal(size=5000)
sy = normal(size=5000)

scatter(sx,sy, linewidth=0, alpha=0.2)

fig = gcf()
id_press = fig.canvas.mpl_connect('button_press_event', on_press)
id_rls = fig.canvas.mpl_connect('button_release_event', on_release)
#id_motion = fig.canvas.mpl_connect('motion_notify_event', on_release)

ax = gca()
print ax
#r = Rectangle(xy=(0,0), height=1, width=2, fill=False, lw=2, alpha=0.2)
#ax.add_artist(r)
#r.set_clip_box(ax.bbox)

show()



"""
A Matplotlib timetrace plot figure with the addition of a QT slider to scroll
the time axis back and forth.
"""

from PySide import QtGui, QtCore
import pylab as plt
import numpy as np

N_SAMPLES = 1e6

def test_plot():
    dx = 1e-3 # this is the sampling period (or grid spacing) of `sample`
    sample = np.random.randn(N_SAMPLES)
    line, = plt.plot([], [], label="Gaussian noise")
    plt.legend(fancybox=True)
    plt.title("Use the slider to scroll and the spin-box to set the width")
    q = ScrollingPlotQT(fig=plt.gcf(), line=line, ydata=sample, dx=1e-3)
    return q   # WARNING: it's important to return this object otherwise
               # python will delete the reference and the GUI will not respond!


class ScrollingPlotQT(object):
    def __init__(self, fig, line, ydata, dx, scroll_step=10):
        # Setup data range variables for scrolling
        self.fig, self.line, self. ydata, self.dx = fig, line, ydata, dx
        self.scroll_step = scroll_step
        self.xmin, self.xmax = 0, dx*ydata.size
        self.plot_width = 1  # axis units
        self.scale = 1e3     # conversion betweeen scrolling units and axis units
        self.disp_points = self.plot_width/self.dx

        # Save some MPL shortcuts
        self.draw = self.fig.canvas.draw
        self.draw_idle = self.fig.canvas.draw_idle
        self.ax = self.fig.axes[0]
        self.set_xlim = self.ax.set_xlim

        # Retrive the QMainWindow used by current figure and add a toolbar
        # to host the new widgets
        QMainWin = fig.canvas.parent()
        toolbar = QtGui.QToolBar(QMainWin)
        QMainWin.addToolBar(QtCore.Qt.BottomToolBarArea, toolbar)

        # Create the slider and spinbox for x-axis scrolling in toolbar
        self.set_slider(toolbar)
        self.set_spinbox(toolbar)

        # Set the initial xlimits coherently with values in slider and spinbox
        self.set_xlim(0, self.plot_width)
        self.ax.set_ylim(ydata.min(), ydata.max())
        
        # Setup the initial plot
        self.line.set_data(np.arange(self.disp_points)*dx,
                ydata[:self.disp_points])

        text0 = self.ax.text(0.01,0.02, "T = ", transform=fig.transFigure)
        self.text = self.ax.text(0.05,0.02, "0", transform=fig.transFigure)
        self.draw()

    def set_slider(self, parent):
        self.slider = QtGui.QSlider(QtCore.Qt.Horizontal, parent=parent)
        self.slider.setTickPosition(QtGui.QSlider.TicksAbove)
        self.slider.setTickInterval((self.xmax-self.xmin)/10.*self.scale)
        self.slider.setMinimum(self.xmin*self.scale)
        self.slider.setMaximum((self.xmax-self.plot_width)*self.scale)
        self.slider.setSingleStep(self.plot_width*self.scale/4.)
        self.slider.setPageStep(self.scroll_step*self.plot_width*self.scale)
        self.slider.setValue(0)  # set the initial position
        self.slider.valueChanged.connect(self.xpos_changed)
        parent.addWidget(self.slider)

    def set_spinbox(self, parent):
        self.spinb = QtGui.QDoubleSpinBox(parent=parent)
        self.spinb.setDecimals(3)
        self.spinb.setRange(0.001, 3600.)
        self.spinb.setSuffix(" s")
        self.spinb.setValue(self.plot_width)   # set the initial width
        self.spinb.valueChanged.connect(self.xwidth_changed)
        parent.addWidget(self.spinb)

    def xpos_changed(self, pos):
        #print("Position (in scroll units) %f\n" %pos)
        pos /= (self.dx*self.scale)  # pos converted in index units for ydata
        self.line.set_ydata(self.ydata[pos:pos+self.disp_points])
        self.text.set_text(pos*self.dx)
        self.draw()

    def xwidth_changed(self, xwidth):
        #pprint("Width (axis units) %f\n" % xwidth)
        if xwidth <= 0: return
        self.plot_width = xwidth
        self.ax.set_xlim(0, self.plot_width)
        self.disp_points = self.plot_width/self.dx
        self.line.set_xdata(np.arange(self.disp_points)*self.dx)
        self.slider.setMaximum((self.xmax-self.plot_width)*self.scale)
        self.slider.setSingleStep(self.plot_width*self.scale/5.)
        self.slider.setPageStep(self.plot_width*self.scale)
        self.xpos_changed(self.slider.value()*self.scale)


if __name__ == "__main__":
    q = test_plot()
    plt.show()



from matplotlib.patches import Rectangle, Ellipse

class GUI_Selection:
    id_motion = 0
    xs = 0
    ys = 0
    xe = 0
    ye = 0
    ax = 0
    pressed = False

GSel = GUI_Selection()

class Point_Selection:
    """Select a point in a plot"""
    xp = 0
    yp = 0
    AX = []
    d = None
    connected = False

PSel = Point_Selection()

def on_press_point(event):
    axb = r_[[ax == event.inaxes for ax in PSel.AX]]
    if not axb.any(): return
    ich = find(axb)
    ax = PSel.AX[ich]

    PSel.xp, PSel.yp, PSel.ich = event.xdata, event.ydata, ich
    #print 'PRESS button=%d, x=%d, y=%d, xdata=%f, ydata=%f' % (
    #        event.button, event.x, event.y, event.xdata, event.ydata)
    pprint("%s %s %s\n" %(PSel.xp, PSel.yp, PSel.ich))
    mburst = PSel.d.mburst[ich]
    t_clk = PSel.xp/PSel.d.clk_p
    mask = (b_start(mburst) < t_clk)*(b_end(mburst) > t_clk)    
    if mask.any():
        ib = find(mask)[0]
        ts = b_start(mburst)[ib]*PSel.d.clk_p
        skew = bleaching1(PSel.d, ich, ib, use_median=True, normalize=True) 
        pprint("Burst [%d-CH%d]: t = %d us   nt = %5.1f   E = %4.2f   Skew = %.2f\n" %\
            (ib, ich+1, ts*1e6, PSel.d.nt[ich][ib], PSel.d.E[ich][ib], skew))
        
def on_press(event):
    if event.inaxes != GSel.ax: return
    GSel.pressed = True
    GSel.xs, GSel.ys = event.xdata, event.ydata
    #print 'PRESS button=%d, x=%d, y=%d, xdata=%f, ydata=%f' % (
    #        event.button, event.x, event.y, event.xdata, event.ydata)
    if hasattr(GSel, 'r'):
        GSel.r.set_height(0); GSel.r.set_width(0)
        GSel.r.set_xy((GSel.xs,GSel.ys))
        GSel.e.height = 0; GSel.e.width = 0
        GSel.e.center = (GSel.xs,GSel.ys)
    else:
        GSel.r = Rectangle(xy=(GSel.xs,GSel.ys), height=0, width=0, 
                fill=False, lw=2, alpha=0.6, color='blue')
        GSel.e = Ellipse(xy=(GSel.xs,GSel.ys), height=0, width=0, 
                fill=False, lw=2, alpha=0.6, color='blue')
        GSel.ax.add_artist(GSel.r)
        GSel.r.set_clip_box(GSel.ax.bbox); GSel.r.set_zorder(10)
        GSel.ax.add_artist(GSel.e)
        GSel.e.set_clip_box(GSel.ax.bbox); GSel.e.set_zorder(10)
    GSel.fig.canvas.draw()
    GSel.id_motion = GSel.fig.canvas.mpl_connect('motion_notify_event', 
            on_motion)

def on_release(event):
    if not GSel.pressed: return
    GSel.pressed = False
    #print 'RELEASE button=%d, x=%d, y=%d, xdata=%f, ydata=%f' % (
    #        event.button, event.x, event.y, event.xdata, event.ydata)
    GSel.fig.canvas.mpl_disconnect(GSel.id_motion)
    #print "Selection: \nE1=%.3f; E2=%.3f; S1=%.3f; S2=%.3f" % \
    #        (GSel.xs, GSel.xe, GSel.ys, GSel.ye)
    E1, E2 = min((GSel.xs, GSel.xe)), max((GSel.xs, GSel.xe))
    S1, S2 = min((GSel.ys, GSel.ye)), max((GSel.ys, GSel.ye))
    print "Selection: \nE1=%.2f, E2=%.2f, S1=%.2f, S2=%.2f" % (E1,E2,S1,S2)

def on_motion(event):
    if event.inaxes != GSel.ax: return
    #print 'MOTION x=%d, y=%d, xdata=%f, ydata=%f' % (
    #        event.x, event.y, event.xdata, event.ydata)
    GSel.xe, GSel.ye = event.xdata, event.ydata
    GSel.r.set_height(GSel.ye-GSel.ys); GSel.r.set_width(GSel.xe-GSel.xs)
    GSel.e.height=(GSel.ye-GSel.ys); GSel.e.width = (GSel.xe-GSel.xs)
    GSel.e.center = (mean([GSel.xs,GSel.xe]),mean([GSel.ys,GSel.ye]))
    GSel.fig.canvas.draw()






# Author: Paul David Harris
# Purpose: Unit tests for burstlib_ext.py functions
# Created: 13 Sept 2022
"""
Unit tests for the burstlib_ext module (bext)

Currently mostly just smoke tests

Running tests requires pytest
"""

from collections import namedtuple
from itertools import product
import pytest
import numpy as np


try:
    import matplotlib
except ImportError:
    has_matplotlib = False  # OK to run tests without matplotlib
else:
    has_matplotlib = True
    matplotlib.use('Agg')  # but if matplotlib is installed, use Agg

# try:
#     import numba
# except ImportError:
#     has_numba = False
# else:
#     has_numba = True


import fretbursts.background as bg
import fretbursts.burstlib as bl
import fretbursts.burstlib_ext as bext
from fretbursts.burstlib import Data
from fretbursts import loader
from fretbursts import select_bursts
from fretbursts.ph_sel import Ph_sel
from fretbursts.phtools import phrates
if has_matplotlib:
    import fretbursts.burst_plot as bplt


# data subdir in the notebook folder
DATASETS_DIR = u'../notebooks/data/'


def _alex_process(d):
    loader.alex_apply_period(d)
    d.calc_bg(bg.exp_fit, time_s=30, tail_min_us=300)
    d.burst_search(L=10, m=10, F=7)

def load_dataset_1ch(process=True):
    fn = "0023uLRpitc_NTP_20dT_0.5GndCl.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    if process:
        _alex_process(d)
    return d

def load_dataset_1ch_nsalex(process=True):
    fn = "dsdna_d7_d17_50_50_1.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    if process:
        _alex_process(d)
    return d

def load_dataset_8ch():
    fn = "12d_New_30p_320mW_steer_3.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    d.calc_bg(bg.exp_fit, time_s=30, tail_min_us=300)
    d.burst_search(L=10, m=10, F=7)
    return d

def load_fake_pax():
    fn = "0023uLRpitc_NTP_20dT_0.5GndCl.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    d.add(ALEX=False, meas_type='PAX')
    loader.alex_apply_period(d)
    d.calc_bg(bg.exp_fit, time_s=30, tail_min_us='auto')
    d.burst_search(L=10, m=10, F=6, pax=True)
    return d

def test_load_group():
    """Smoke test to see if loading groupd data works"""
    fn = ['HP3_TE150_SPC630.hdf5', 'HP3_TE200_SPC630.hdf5', 'HP3_TE250_SPC630.hdf5', 'HP3_TE300_SPC630.hdf5']
    fn = [DATASETS_DIR + f for f in fn]
    d = loader.photon_hdf5(fn)
    assert d.nch == len(fn)
    
def load_dataset_grouped(process=True):
    fn = ['HP3_TE150_SPC630.hdf5', 'HP3_TE200_SPC630.hdf5', 'HP3_TE250_SPC630.hdf5', 'HP3_TE300_SPC630.hdf5']
    fn = [DATASETS_DIR + f for f in fn]
    d = loader.photon_hdf5(fn)
    if process:
        _alex_process(d)
    return d


@pytest.fixture(scope="module")
def data_8ch(request):
    d = load_dataset_8ch()
    return d

@pytest.fixture(scope="module")
def data_1ch(request):
    d = load_dataset_1ch()
    return d

@pytest.fixture(scope="module")
def data_1ch_nsalex(request):
    d = load_dataset_1ch_nsalex()
    return d

@pytest.fixture(scope="module")
def data_grouped(request):
    d = load_dataset_grouped()
    return d

@pytest.fixture(scope="module", params=[
                                    load_dataset_1ch,
                                    load_dataset_1ch_nsalex,
                                    load_dataset_8ch,
                                    load_dataset_grouped
                                    ])
def data(request):
    load_func = request.param
    d = load_func()
    return d

@pytest.mark.parametrize("data_ch, process", product([load_dataset_1ch, 
                                             load_dataset_1ch_nsalex, 
                                             load_dataset_8ch],
                                            [True, False]))
def test_group_data(data_ch, process):
    if data_ch is load_dataset_8ch:
        orig_d = data_ch()
    else:
        orig_d = data_ch(process=process)
    n = orig_d.nch
    d = bext.group_data([orig_d for _ in range(8)])
    assert n*8 == d.nch
    for field in Data.ph_fields:
        if hasattr(d, field):
            assert len(d[field]) == 8*n, f"{field} not correct length"
    if not process and orig_d.alternated:
        loader.alex_apply_period(d)
        for field in Data.ph_fields:
            if hasattr(d, field):
                assert len(d[field]) == 8*n, f"{field} not correct length"
    d.calc_bg(bg.exp_fit, time_s=30, tail_min_us='auto')
    d.burst_search(m=10, F=7, L=10)
    for field in Data.burst_fields:
        if hasattr(d, field):
            assert len(d[field]) == 8*n, f'{field} not correct length'

def test_join_data(data):
    """Smoke test for bext.join_data() function.
    """
    d = data
    dj = bext.join_data([d, d.copy()])
    assert (dj.num_bursts == 2 * d.num_bursts).all()
    for bursts in dj.mburst:
        assert (np.diff(bursts.start) > 0).all()

def test_burst_search_and_gate(data_1ch):
    """Test consistency of burst search and gate."""
    d = data_1ch
    assert d.alternated

    # Smoke tests
    bext.burst_search_and_gate(d, F=(6, 8))
    bext.burst_search_and_gate(d, m=(12, 8))
    bext.burst_search_and_gate(d, min_rate_cps=(60e3, 40e3))
    if d.nch > 1:
        mr1 = 35e3 + np.arange(d.nch) * 1e3
        mr2 = 30e3 + np.arange(d.nch) * 1e3
        bext.burst_search_and_gate(d, min_rate_cps=(mr1, mr2))

    # Consistency test
    d_dex = d.copy()
    d_dex.burst_search(ph_sel=Ph_sel(Dex='DAem'))
    d_aex = d.copy()
    d_aex.burst_search(ph_sel=Ph_sel(Aex='Aem'))
    d_and = bext.burst_search_and_gate(d)
    for bursts_dex, bursts_aex, bursts_and, ph in zip(
            d_dex.mburst, d_aex.mburst, d_and.mburst, d.iter_ph_times()):
        ph_b_mask_dex = bl.ph_in_bursts_mask(ph.size, bursts_dex)
        ph_b_mask_aex = bl.ph_in_bursts_mask(ph.size, bursts_aex)
        ph_b_mask_and = bl.ph_in_bursts_mask(ph.size, bursts_and)
        assert (ph_b_mask_and == ph_b_mask_dex * ph_b_mask_aex).all()

def test_burst_data(data):
    """Test for bext.burst_data()"""
    bext.burst_data(data, include_bg=True, include_ph_index=True)
    bext.burst_data(data, include_bg=False, include_ph_index=True)
    bext.burst_data(data, include_bg=True, include_ph_index=False)
    bext.burst_data(data, include_bg=False, include_ph_index=False)




def test_asymmetry(data):
    d = data
    for i in range(data.nch):
        asym = bext.asymmetry(data, i, dropnan=False)
        assert len(asym) == d.mburst[i].size
        if np.any(np.isnan(asym)):
            nan_count = np.isnan(asym).sum()
            asym = bext.asymmetry(d, i)
            assert len(asym) == d.mburst[i].size - nan_count


def test_calc_mdelays_hist(data):
    """Smoke test for calc_mdelays_hist"""
    d = data
    for i in range(d.nch):
        for ph_sel in [Ph_sel('all'), Ph_sel(Dex='Dem'), Ph_sel(Dex='Aem')]:
            bext.calc_mdelays_hist(d, ich=i)

def test_burst_fitter(data):
    d = data
    bext.bursts_fitter(d)
    assert hasattr(d, 'E_fitter')
    if d.alternated:
        bext.bursts_fitter(d, burst_data='S')
        assert hasattr(d, 'S_fitter')



#
# FRETBursts - A single-molecule FRET burst analysis toolkit.
#
# Copyright (C) 2014 Antonino Ingargiola <tritemio@gmail.com>
#
"""
Misc utility functions
"""

import os
import sys
import numpy as np


def selection_mask(arr, values):
    """Return a boolean mask, True when `arr` is one element of `values`.

    This function generalizes the comparison `arr == values`
    where `values` can be a scalar or a sequence.
    If a sequence, the returned mask is True each time `arr` has an element
    in `values`.

    Arguments:
        arr (array): input array for which a mask is computed.
        values (scalar or sequence): one or more values to be selected in
            `arr`.
    
    Returns:
        Boolean mask same size as `arr`. True where `arr` has an element 
        in `values`.
    """
    values = np.atleast_1d(values)
    mask = arr == values[0]
    for value in values[1:]:
        mask += arr == value
    return mask



def _is_list_of_arrays(obj):
    return isinstance(obj, list) and np.all([isinstance(v, np.ndarray)
                                             for v in obj])

class HistData(object):
    """Stores histogram counts and bins and provides derived fields.

    Attributes:
        counts (array, ints): array of counts in each bin
        bins (array): array of bin edges. Size is size(counts) + 1.
        bincenters (array): array of bin  centers. Size is size(counts).
        pdf (array, floats): array of normalized counts (aka PDF)
    """
    def __init__(self, counts, bins):
        self.counts = counts
        self.bins = bins
        self.binwidth = bins[1] - bins[0]

    @property
    def bincenters(self):
        if not hasattr(self, '_bincenters'):
            self._bincenters = self.bins[:-1] + 0.5*self.binwidth
        return self._bincenters

    @property
    def pdf(self):
        if not hasattr(self, '_pdf'):
            self._pdf = np.array(self.counts, dtype=np.float64)
            self._pdf /= (self.counts.sum() * self.binwidth)
        return self._pdf


def clk_to_s(t_ck, clk_p=12.5*1e-9):
    """Convert clock cycles to seconds."""
    return t_ck*clk_p

def pprint(s, mute=False):
    """Print immediately, even if inside a busy loop."""
    if mute: return
    sys.stdout.write(s)
    sys.stdout.flush()

def deprecate(function, old_name, new_name):
    def deprecated_function(*args, **kwargs):
        pprint("Function name %s is deprecated, use %s instead.\n" %\
                (old_name, new_name))
        res = function(*args, **kwargs)
        return res
    return deprecated_function

def shorten_fname(f):
    """Return a path with only the last subfolder (i.e. measurement date)."""
    return '/'.join(f.split('/')[-2:])

def binning(times, bin_width_ms=1, max_num_bins=1e5, clk_p=12.5e-9):
    """Return the binned histogram of array times."""
    bin_width_clk = (bin_width_ms*1e-3)/clk_p
    num_bins = min(times.max()/bin_width_clk, max_num_bins)
    h = np.histogram(times[times<(num_bins*bin_width_clk)], bins=num_bins)
    return h


def mkdir_p(path):
    """Create the path if not existent, otherwise do nothing.
    If `path` exists, and is not a dir, raise an exception.
    """
    import errno
    try:
        os.makedirs(path)
    except OSError as exc: # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise exc

def download_file(url, save_dir='./'):
    """Download a file from `url` saving it to disk.

    The file name is taken from `url` and left unchanged.
    The destination dir can be set using `save_dir`
    (Default: the current dir).
    """
    # Check if local path already exist
    fname = url.split('/')[-1]
    print('URL:  %s' % url)
    print('File: %s\n ' % fname)

    path = '/'.join([os.path.abspath(save_dir), fname])
    if os.path.exists(path):
        print('File already on disk: %s \nDelete it to re-download.' % path)
        return

    from urllib.request import urlopen, urlretrieve
    from urllib.error import HTTPError, URLError

    # Check if the URL is valid
    try:
        urlopen(url)
    except URLError as e:
        print('Wrong URL or no connection.\n\nError:\n%s\n' % e)
    except HTTPError:
        print('URL not found: ' + url)
        return

    # Download the file
    def _report(blocknr, blocksize, size):
        current = blocknr*blocksize/2**20
        sys.stdout.write(
            "\rDownloaded {0:4.1f} / {1:4.1f} MB".format(current, size/2**20))
    mkdir_p(save_dir)
    urlretrieve(url, path, _report)

def _large_equal(val0, val1):
    if type(val0) != type(val1):
        return False
    elif isinstance(val0, np.ndarray):
        if np.any(val0.shape != val1.shape):
            return False
        else:
            return np.all(val0 == val1)
    elif isinstance(val0, (list, tuple)):
        if len(val0) == len(val1):
            return np.all([_large_equal(v0, v1) for v0, v1 in zip(val0, val1)])
        else:
            return False
    else:
        return val0 == val1


def dict_equal(*dicts):
    key0 = dicts[0].keys()
    comp = np.all([key0 == dct.keys() for dct in dicts])
    if comp:
        for key, val0 in dicts[0].items():
            for dct in dicts[1:]:
                if not _large_equal(val0, dct[key]):
                    comp = False
                    break 
            if not comp:
                break
    return comp


# Readthedocs default blue theme
html_style = None

# Mock modules that are not present on readthedocs
# http://read-the-docs.readthedocs.org/en/latest/faq.html#i-get-import-errors-on-libraries-that-depend-on-c-modules
import sys
class Mock(object):
    def __init__(self, *args, **kwargs):
        self._mock = True   # if this exists the object is a mock
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr__(cls, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        elif name[0] == name[0].upper():
            mockType = type(name, (), {})
            mockType.__module__ = __name__
            return mockType
        else:
            return Mock()


MOCK_MODULES = [
                 #'numpy', 'numpy.random',
                 #'scipy', 'scipy.stats', 'scipy.optimize', 'scipy.special',
                 #'scipy.ndimage', 'scipy.interpolate',
                 #'builtins',
                 #'matplotlib', 'matplotlib.pyplot', 'matplotlib.mlab',
                 #'matplotlib.patches', 'matplotlib.collections',
                 #'mpl_toolkits.axes_grid1', 'seaborn', 'lmfit',
                 #'PySide','PySide.QtCore','PySide.QtGui',
                 #'tables', 'pandas',
                 #'phconvert', 'phconvert.hdf5', 'phconvert.loader',
                 ]

for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()



# encoding: utf-8  μ αβγδ εζηθ κλμν ξοπρ ςστυ φχψω

##
# Experiments with fonts
#
# NOTE: to change default font in realtime change font.family to your
# font name. Keeping the family to sans-serif and changing font.sans-serif
# change the fornt only the first time!!!
#

font1 = {'fontname':'Liberation Sans','fontsize':16} # best general font
font2 = {'fontname':'Arial','fontsize':16} # Best for numbers, worst greek
font3 = {'fontname':'Dejavu Sans','fontsize':16}
font4 = {'fontname':'Ubuntu','fontsize':16} # Nice 1 number very sim. to Arial
font5 = {'fontname':'cmex10', 'fontsize':16}
font5 = {'fontname':'Computer Modern Sans serif', 'fontsize':16}

rcParams["font.family"] = font5['fontname']

greek = u"ταβγδεζηθικλμνξοπρςστυφχψω"
greek_math_it = u"𝛼𝛽𝛾𝛿𝜀𝜁𝜂𝜃𝜆𝜇𝜈𝜉𝜋𝜌𝜍𝜎𝜏𝜐𝜑𝜒𝜓𝜔𝜕𝜖𝜗𝜘𝜙𝜚" # apparently does not work
s = " 0123456789 "

figure()
title(s) # -> uses default font

# Use custom fonts
text(0.1,0.9, font1['fontname']+s+greek, **font1) 
text(0.1,0.7, font2['fontname']+s+greek, **font2)
text(0.1,0.3, font3['fontname']+s+greek, **font3)
text(0.1,0.1, font4['fontname']+s+greek, **font4)
text(0.1,0.5, font5['fontname']+s+greek, **font4)

## Last (default) setting before show() takes effect
#rcParams["font.family"] = font1['fontname']

show()




#!/usr/bin/env python
#
# Demonstration file: how to embed a matplotlib figure in a QT4 window
#

import sys
from PyQt4 import QtGui, QtCore

from matplotlib.figure import Figure

# Import from matplotlib the FigureCanvas with QTAgg backend
from matplotlib.backends.backend_qt4agg \
        import FigureCanvasQTAgg as FigureCanvas

# Import the matplotlib Toolbar2
from matplotlib.backends.backend_qt4agg \
        import NavigationToolbar2QT as NavigationToolbar

class PlotWindow(QtGui.QWidget):
    def __init__(self, show=True, create_axis=False):
        super(PlotWindow, self).__init__()
        
        # Create the plot canvas ...
        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)        
        if create_axis: self.axis = self.figure.add_subplot(111)
        #self.canvas.setParent(self)
        
        # ... and the toolbar
        self.toolbar = NavigationToolbar(self.canvas, self)

        # Pack all in QVBoxLayout
        self.vbox = QtGui.QVBoxLayout(self)
        self.vbox.addWidget(self.canvas)
        self.vbox.addWidget(self.toolbar)

        if show: self.show()

class ScrollPlotWindow(PlotWindow):
    def __init__(self):
        super(ScrollPlotWindow, self).__init__(show=False, create_axis=True)
        self.slider = QtGui.QSlider(QtCore.Qt.Horizontal, self)
        self.slider.valueChanged.connect(self.changeValue)
        self.slider.setTickPosition(QtGui.QSlider.TicksBelow)
        self.slider.setTickInterval(10)
        self.vbox.insertWidget(1, self.slider)
        self.show()

    def changeValue(self, value):
        self.axis.set_xlim(value,value+1)
        self.canvas.draw()

def main():
    #win = PlotWindow()
    win = ScrollPlotWindow()
    #win = TestWindow()
    sys.exit(app.exec_())    

if __name__ == '__main__':
    app = QtGui.QApplication(sys.argv)
    #pass
    main()    



"""
Example on how to add widgets the toolbar of a Matplotlib figure using the
QT backend.

No QT application is created, only the toolbar of the native MPL figure is
modified.
"""


from PySide import QtGui, QtCore
import matplotlib


def test():
    plot([1,2,3], lw=2)
    q = qt4_interface(gcf())
    return q   # WARNING: it's paramount to return the object otherwise, with 
               # no references, python deletes it and the GUI doesn't respond!
    
class qt4_interface:
    def __init__(self,fig):
        self.fig = fig
        toolbar = fig.canvas.toolbar
        
        self.line_edit = QtGui.QLineEdit()
        toolbar.addWidget(self.line_edit)
        self.line_edit.editingFinished.connect(self.do_something)

        self.spinbox = QtGui.QDoubleSpinBox()
        toolbar.addWidget(self.spinbox)
        self.spinbox.valueChanged.connect(self.do_something2)

    def do_something(self, *args):
        self.fig.axes[0].set_title(self.line_edit.text())
        self.fig.canvas.draw()
        #f = open('l','a'); f.write('yes\n'); f.flush(); f.close()    
    
    def do_something2(self, *args):
        self.fig.axes[0].set_xlim(0, self.spinbox.value())
        self.fig.canvas.draw()
        #f = open('l','a'); f.write('yes\n'); f.flush(); f.close()



# Author: Paul David Harris
# Purpose: Unit tests for burst_plot.py functions
# Created: 13 Sept 2022
"""
Tests of plotting functions

These are mostly smoke tests
"""

from collections import namedtuple
from itertools import product
import pytest
import numpy as np


import matplotlib
matplotlib.use('Agg')  # but if matplotlib is installed, use Agg
import matplotlib.pyplot as plt
# try:
#     import numba
# except ImportError:
#     has_numba = False
# else:
#     has_numba = True


import fretbursts.background as bg
import fretbursts.burstlib as bl
import fretbursts.burstlib_ext as bext
from fretbursts.burstlib import Data
from fretbursts import loader
from fretbursts import select_bursts
from fretbursts.ph_sel import Ph_sel
from fretbursts.phtools import phrates
import fretbursts.burst_plot as bplt


# data subdir in the notebook folder
DATASETS_DIR = u'../notebooks/data/'


def _alex_process(d):
    loader.alex_apply_period(d)
    d.calc_bg(bg.exp_fit, time_s=30, tail_min_us=300)
    d.burst_search(L=10, m=10, F=7)

def load_dataset_1ch(process=True):
    fn = "0023uLRpitc_NTP_20dT_0.5GndCl.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    if process:
        _alex_process(d)
    return d

def load_dataset_1ch_nsalex(process=True):
    fn = "HP3_TE150_SPC630.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    if process:
        _alex_process(d)
    return d

def load_dataset_8ch():
    fn = "12d_New_30p_320mW_steer_3.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    d.calc_bg(bg.exp_fit, time_s=30, tail_min_us=300)
    d.burst_search(L=10, m=10, F=7)
    return d

def load_fake_pax():
    fn = "0023uLRpitc_NTP_20dT_0.5GndCl.hdf5"
    fname = DATASETS_DIR + fn
    d = loader.photon_hdf5(fname)
    d.add(ALEX=False, meas_type='PAX')
    loader.alex_apply_period(d)
    d.calc_bg(bg.exp_fit, time_s=30, tail_min_us='auto')
    d.burst_search(L=10, m=10, F=6, pax=True)
    return d

    
def load_dataset_grouped(process=True):
    fn = ['HP3_TE150_SPC630.hdf5', 'HP3_TE200_SPC630.hdf5', 'HP3_TE250_SPC630.hdf5', 'HP3_TE300_SPC630.hdf5']
    fn = [DATASETS_DIR + f for f in fn]
    d = loader.photon_hdf5(fn)
    if process:
        _alex_process(d)
    return d


@pytest.fixture(scope="module")
def data_8ch(request):
    d = load_dataset_8ch()
    return d

@pytest.fixture(scope="module")
def data_1ch(request):
    d = load_dataset_1ch()
    return d

@pytest.fixture(scope="module")
def data_1ch_nsalex(request):
    d = load_dataset_1ch_nsalex()
    return d

@pytest.fixture(scope="module")
def data_grouped(request):
    d = load_dataset_grouped()
    return d

@pytest.fixture(scope="module", params=[
                                    load_dataset_1ch,
                                    load_dataset_1ch_nsalex,
                                    load_dataset_8ch,
                                    load_dataset_grouped
                                    ])
def data(request):
    load_func = request.param
    d = load_func()
    return d

@pytest.fixture(scope="module", params=[
                                    load_dataset_8ch,
                                    load_dataset_grouped
                                    ])
def data_mch(request):
    load_func = request.param
    d = load_func()
    return d

@pytest.fixture(scope='module', params=[load_dataset_1ch,
                                        load_dataset_1ch_nsalex,
                                        load_dataset_grouped])
def data_alex(request):
    load_func = request.param
    d = load_func()
    return d

##
#  Multi-channel plot functions
#

def test_mch_plot_bg(data_mch):
    d = data_mch
    bplt.mch_plot_bg(d)
    plt.close()

def test_mch_plot_bg_ratio(data_mch):
    d = data_mch
    bplt.mch_plot_bg_ratio(d)
    plt.close()

def test_mch_plot_bsize(data_mch):
    d = data_mch
    bplt.mch_plot_bsize(d)
    plt.close()

##
#  Timetrace plots
#
@pytest.fixture(scope='module', params = (bplt.timetrace_single, bplt.ratetrace_single,
                                          bplt.hist_interphoton_single))
def ratetraces(request):
    return request.param

def test_trace_single(data, ratetraces):
    """Test plotting functions for time traces that do so on single Ph_sel"""
    d = data
    ph_sel_alex = (Ph_sel('all'), Ph_sel(Dex='Dem'), Ph_sel(Dex='Aem'), Ph_sel(Aex='Dem'), Ph_sel(Aex='Aem'))
    ph_sel_single = (Ph_sel('all'), Ph_sel(Dex='Dem'), Ph_sel(Dex='Aem'))
    ph_sel_list = ph_sel_alex if d.alternated else ph_sel_single
    for i in range(d.nch):
        for ph_sel in ph_sel_list:
            bplt.dplot(d, ratetraces, i=i, ph_sel=ph_sel)
    plt.close()

@pytest.fixture(scope='module', params = (bplt.timetrace, bplt.ratetrace, 
                                          bplt.timetrace_bg, bplt.timetrace_fret,
                                          bplt.timetrace_fret_scatter, bplt.time_ph))
def timetraces(request):
    return request.param

def test_trace(data, timetraces):
    """Test general time trace type functions"""
    d = data
    for i in range(d.nch):
        bplt.dplot(d, timetraces, i=i)
    plt.close()


@pytest.fixture(scope='module', params = (bplt.hist_size, bplt.hist_width,
                                          bplt.hist_brightness, bplt.hist_size_all,
                                          bplt.hist_fret, bplt.hist_interphoton,
                                          bplt.hist_ph_delays, bplt.hist_mdelays,
                                          bplt.hist_mrates, bplt.hist_sbr, 
                                          bplt.hist_burst_phrate, bplt.hist_burst_delays,
                                          bplt.hist_asymmetry))
def hists(request):
    return request.param

def test_hist(data, hists):
    d = data
    bplt.dplot(d, hists, i=None)
    for i in range(d.nch):
        bplt.dplot(d, hists, i=i)
    plt.close()

def test_hist_S(data_alex):
    d = data_alex
    bplt.dplot(d, bplt.hist_S, i=None)
    for i in range(d.nch):
        bplt.dplot(d, bplt.hist_S)
    plt.close()

@pytest.fixture(scope='module', params = (bplt.hist2d_alex, bplt.hexbin_alex,
                                          bplt.scatter_alex, bplt.scatter_naa_nt))
def ES_plots(request):
    return request.param

def test_ES_plots(data_alex, ES_plots):
    d = data_alex
    bplt.dplot(d, ES_plots, i=None)
    if ES_plots in (bplt.scatter_alex, bplt.scatter_naa_nt):
        bplt.dplot(d, ES_plots, i=0, color_style='kde')
    for i in range(d.nch):
        bplt.dplot(d, ES_plots, i=i)
    plt.close()

@pytest.fixture(scope='module', params = (bplt.scatter_width_size, bplt.scatter_rate_da,
                                          bplt.scatter_fret_size, bplt.scatter_fret_nd_na,
                                          bplt.scatter_fret_width, bplt.scatter_da,))
def scatterplots(request):
    return request.param

def test_scatterplots(data, scatterplots):
    d = data
    bplt.dplot(d, scatterplots, i=None)
    for i in range(d.nch):
        bplt.dplot(d, scatterplots, i=i)
    plt.close()



from setuptools import setup
from setuptools.extension import Extension
import numpy as np

## Metadata
project_name = 'fretbursts'

from Cython.Build import cythonize
ext_modules = [Extension("fretbursts.burstsearch_c",
                         [project_name + "/phtools/burstsearch_c.pyx"]),
               Extension("fretbursts.phrates_c",
                         [project_name + "/phtools/phrates_cy.pyx"],
                          include_dirs = ["."],)]

## Configure setup.py commands


setup(
      include_dirs = [np.get_include()],
      ext_modules = cythonize(ext_modules),
      include_package_data = True,
      packages = ['fretbursts', 'fretbursts.utils', 'fretbursts.fit',
                  'fretbursts.phtools', 'fretbursts.dataload'],
      )



"""
A Matplotlib timetrace plot figure with the addition of a QT slider to scroll
the time axis back and forth.
"""

from PySide import QtGui, QtCore
import pylab as plt
import numpy as np

N_SAMPLES = 1e6

def test_plot():
    time = np.arange(N_SAMPLES)*1e-3
    sample = np.random.randn(N_SAMPLES)
    plt.plot(time, sample, label="Gaussian noise")
    plt.title("1000s Timetrace \n (use the slider to scroll and the spin-box "
              "to set the width)")
    plt.xlabel('Time (s)')
    plt.legend(fancybox=True)
    q = ScrollingToolQT(plt.gcf(), scroll_step=1)
    return q   # WARNING: it's important to return this object otherwise
               # python will delete the reference and the GUI will not respond!


class ScrollingToolQT(object):
    def __init__(self, fig, scroll_step=10):
        # Setup data range variables for scrolling
        self.fig = fig
        self.scroll_step = scroll_step
        self.xmin, self.xmax = fig.axes[0].get_xlim()
        self.width = 1 # axis units
        self.pos = 0   # axis units

        self.scale = 1e3 # conversion betweeen scrolling units and axis units
        
        # Save some MPL shortcuts
        self.draw = self.fig.canvas.draw
        self.draw_idle = self.fig.canvas.draw_idle
        self.ax = self.fig.axes[0]
        self.set_xlim = self.ax.set_xlim

        # Retrive the QMainWindow used by current figure and add a toolbar
        # to host the new widgets
        QMainWin = fig.canvas.parent()
        toolbar = QtGui.QToolBar(QMainWin)
        QMainWin.addToolBar(QtCore.Qt.BottomToolBarArea, toolbar)

        # Create the slider and spinbox for x-axis scrolling in toolbar
        self.set_slider(toolbar)
        self.set_spinbox(toolbar)

        # Set the initial xlimits coherently with values in slider and spinbox
        self.set_xlim(0, self.width)
        self.draw()

    def set_slider(self, parent):
        # Slider only support integer ranges so use ms as base unit
        smin, smax = self.xmin*self.scale, self.xmax*self.scale

        self.slider = QtGui.QSlider(QtCore.Qt.Horizontal, parent=parent)
        self.slider.setTickPosition(QtGui.QSlider.TicksAbove)
        self.slider.setTickInterval((smax-smin)/10.)
        self.slider.setMinimum(smin)
        self.slider.setMaximum(smax-self.width*self.scale)
        self.slider.setSingleStep(self.width*self.scale/5.)
        self.slider.setPageStep(self.width*self.scale)
        self.slider.setValue(0)  # set the initial position
        self.slider.valueChanged.connect(self.xpos_changed)
        parent.addWidget(self.slider)

    def set_spinbox(self, parent):
        self.spinb = QtGui.QDoubleSpinBox(parent=parent)
        self.spinb.setDecimals(3)
        self.spinb.setRange(0.001, 3600.)
        self.spinb.setSuffix(" s")
        self.spinb.setValue(self.width)   # set the initial width
        self.spinb.valueChanged.connect(self.xwidth_changed)
        parent.addWidget(self.spinb)

    def xpos_changed(self, pos):
        #pprint("Position (in scroll units) %f\n" %pos)
        pos /= self.scale
        self.set_xlim(pos, pos + self.width)
        self.draw_idle()

    def xwidth_changed(self, xwidth):
        #pprint("Width (axis units) %f\n" % xwidth)
        if xwidth <= 0: return
        self.width = xwidth
        self.slider.setSingleStep(self.width*self.scale/5.)
        self.slider.setPageStep(self.scroll_step*self.width*self.scale)
        old_xlim = self.ax.get_xlim()
        self.xpos_changed(old_xlim[0] * self.scale)

if __name__ == "__main__":
    q = test_plot()
    plt.show()
